// timedep.cpp
// by Glenn G. Chappell
// October 2003
//
// For CS 381
// Time-Dependent Motion Demo
//
// Draws a moving object that goes at the
//  same speed, regardless of system speed.

#include <iostream>
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <time.h>    // Some versions of MS-Vis C++ have broken <ctime>
//using std::clock;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff, includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400;    // Starting window width & height, in pixels

// Keyboard
const int ESCKEY = 27;           // ASCII value of escape character

// Globals affecting display
bool motionon = false;           // Is object moving?
const double objspeed = 1.0;     // Base object speed in world units per second
double currobjspeed = objspeed;  // Current object speed
const double objsize = 0.15;     // Half of side of object
double objxpos = 0.;             // Object position x-coordinate
const double maxobjxpos = 1.0-objsize;
const double minobjxpos = -1.0+objsize;
                                 // Max & min for above


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// delay
// Given an amount of time, in seconds.
// Pauses for that amount of time.
// Call this function is (say) the display function,
//  in order to slow down the program.
void delay(double delayamount)
{
   double starttime = double(clock())/CLOCKS_PER_SEC;
   double currtime;
   do
   {
      currtime = double(clock())/CLOCKS_PER_SEC;
   }
   while(currtime < starttime + delayamount);
}


// display
// The GLUT display function
void display()
{
   glClear(GL_COLOR_BUFFER_BIT);

   //delay(0.05);  // Uncomment this line to slow down the display,
                   //  thus simulating a very slow system.

   // Draw moving (?) object
   glColor3d(0.9, 0.1, 0.5);
   glBegin(GL_QUADS);
      glVertex2d(objxpos-objsize, -objsize);
      glVertex2d(objxpos+objsize, -objsize);
      glVertex2d(objxpos+objsize,  objsize);
      glVertex2d(objxpos-objsize,  objsize);
   glEnd();

   // Draw text
   glColor3d(0., 0., 0.);
   printbitmap("Time-Dependent Motion Demo", -0.9, 0.9);
   printbitmap("Object should move at nearly", -0.9, 0.8);
   printbitmap("the same speed on any system.", -0.9, 0.7);
   printbitmap("M     Toggle motion", -0.9, 0.6);  // Blank line between previous & this
   printbitmap("Esc   Quit", -0.9, 0.5);

   glutSwapBuffers();
}


// idle
// The GLUT idle function
void idle()
{
   // Do time-dependent motion. We want the speed of moving objects in the
   //  scene to remain nearly the same regardless of processor speed.
   // We use the expression    double(clock())/CLOCKS_PER_SEC
   // which gives the number of seconds since the current process started.
   // Faster processors will produce *smoother* motion at the same
   //  apparent speed.
   static double saved_time = double(clock())/CLOCKS_PER_SEC;
      // Time of last call to idle function.
      // Is static so it retains its value between calls.
      // Is only initialized ONCE.
   double curr_time = double(clock())/CLOCKS_PER_SEC;
      // Current time
   double elapsed_time = curr_time - saved_time;
      // Time since last call to idle function

   if (motionon)
   {
      objxpos += currobjspeed * elapsed_time;
      if (objxpos > maxobjxpos)
      {
         objxpos = 2*maxobjxpos-objxpos;
         currobjspeed = -objspeed;
      }
      if (objxpos < minobjxpos)
      {
         objxpos = 2*minobjxpos-objxpos;
         currobjspeed = objspeed;
      }
      glutPostRedisplay();
   }

   saved_time = curr_time;  // Save the current time for the next idle call

   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   case 'm':  // m - toggle motion
   case 'M':  // upper- and lower-case are different!
      motionon = !motionon;
      glutPostRedisplay();
      break;
   }
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
    glViewport(0, 0, w, h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // We set up coordinate system so that aspect ratios are always correct,
    //  and the region from -1..1 in x & y always just fits in the viewport.
    if (w > h)
    {
       gluOrtho2D(-double(w)/h, double(w)/h, -1., 1.);
    }
    else
    {
       gluOrtho2D( -1., 1., -double(h)/w, double(h)/w);
    }

    glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}


// init
// Initialization
// Called by main after window creation
void init()
{
   glClearColor(1., 1., 1., 0.);
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - Time-Dependent Motion Demo");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display); 
   glutIdleFunc(idle);
   glutKeyboardFunc(keyboard);
   glutReshapeFunc(reshape);

   // Do something
   glutMainLoop();

   return 0;
}
